# from .util import utility, stoppable_thread
# from .printer import printer, terminal_colors, table_printer
# from .progress import progress
# from .file import fileutils
# from .network.downloader import downloader
#
# __all__ = [
#     'stoppable_thread',
#     'utility',
#     'terminal_colors',
#     'printer',
#     'table_printer',
#     'fileutils',
#     'progress',
#     'downloader'
# ]



from .download_event import DownloadEventType, ProgressEventArgs, DownloadStartEvent, DownloadFailureEvent, \
    DownloadCompleteEvent
from .downloader import Downloader
from .downloader_tool import DownloaderTool

__all__ = [
    'DownloadEventType',
    'ProgressEventArgs',
    'DownloadStartEvent',
    'DownloadFailureEvent',
    'DownloadCompleteEvent',
    'Downloader',
    'DownloaderTool'
]



from threading import Thread, Event


class StoppableThread(Thread):
    def __init__(self, target, args=()):
        super(StoppableThread, self).__init__(target=target, args=args)
        self._stop_event = None

    def stop(self) -> None:
        self._stop_event.set()

    def _init_stop_event(self) -> None:
        self._stop_event = Event()

    def _is_stopped(self) -> bool:
        return self._stop_event.is_set()



import typing
import re

from .terminal_colors import TerminalColors


class Printer:
    _newline: bool = True

    @staticmethod
    def apply_background(message, bg_color):
        end_char = '' if Printer._formatted(message) else TerminalColors.END_CHAR
        msg = str.format("{}{}{}", bg_color, message, end_char)
        return msg

    @staticmethod
    def format(text: str, color: str = None, bold: bool = False, underlined: bool = False) -> str:
        """Formats a string with the specified color, boldness, and underlining."""
        color = color if color is not None else ''
        bolded = TerminalColors.BOLD if bold else ''
        underlined = TerminalColors.UNDERLINE if underlined else ''
        end = TerminalColors.END_CHAR
        return f'{color}{bolded}{underlined}{text}{end}'

    @staticmethod
    def format_hex(text: str, text_hex: str, bg_hex: str | None = None,
                   bold: bool = False, underlined: bool = False) -> str:
        """Formats a string with the specified color, boldness, and underlining."""
        if bg_hex is None:
            text_rgb: tuple = Printer.hex_to_rgb(text_hex)
            return Printer.format_rgb(text, text_rgb, None, bold, underlined)
        else:
            text_rgb: tuple = Printer.hex_to_rgb(text_hex)
            bg_rgb: tuple = Printer.hex_to_rgb(bg_hex)
            return Printer.format_rgb(text, text_rgb, bg_rgb, bold, underlined)

    @staticmethod
    def format_rgb(text: str, text_rgb: tuple[int, int, int], bg_rgb: tuple[int, int, int] | None = None,
                   bold: bool = False, underlined: bool = False) -> str:
        """Formats a string with the specified color, boldness, and underlining."""
        bolded = TerminalColors.BOLD if bold else ''
        underlined = TerminalColors.UNDERLINE if underlined else ''
        end = TerminalColors.END_CHAR
        if bg_rgb is None:
            return f'\033[38;2;{text_rgb[0]};{text_rgb[1]};{text_rgb[2]}m{bolded}{underlined}{text}{end}'
        else:
            return f'\033[38;2;{text_rgb[0]};{text_rgb[1]};{text_rgb[2]}m' \
                   f'\033[48;2;{bg_rgb[0]};{bg_rgb[1]};{bg_rgb[2]}m{bolded}{underlined}{text}{end}'

    @staticmethod
    def error(text: str, bold: bool = False, underlined: bool = False) -> None:
        """Prints an error message to the console."""
        Printer.print_hex(text, TerminalColors.ERROR, bold=bold, underlined=underlined)

    @staticmethod
    def get_color_string(color: str | tuple[int, int, int]):
        if isinstance(color, tuple):
            return Printer.rgb_to_hex(color)
        else:
            return color

    @staticmethod
    def hex_to_rgb(hex_color: str) -> tuple[int, int, int]:
        """Converts a hex string to an RGB tuple."""
        hex_without_hash = hex_color.replace('#', '') if hex_color.startswith('#') else hex_color
        return typing.cast(tuple[int, int, int], tuple(int(hex_without_hash[i:i + 2], 16) for i in (0, 2, 4)))

    @staticmethod
    def info(text: str, bold: bool = False, underlined: bool = False) -> None:
        """Prints an info message to the console."""
        Printer.print_hex(text, TerminalColors.INFO, bold=bold, underlined=underlined)

    @staticmethod
    def prevent_newline(prevent: bool = True) -> None:
        """Prevents a newline from being printed to the console."""
        if Printer._newline != prevent:
            return
        Printer._newline = not prevent
        if Printer._newline:
            print()

    @staticmethod
    def print_hex(text: str, text_hex: str, bg_hex: str | None = None,
                  bold: bool = False, underlined: bool = False) -> None:
        """Prints a message to the console."""
        rgb: tuple = Printer.hex_to_rgb(text_hex)
        bg_rgb: tuple = Printer.hex_to_rgb(bg_hex) if bg_hex is not None else None
        Printer.print_rgb(text, rgb, bg_rgb, bold, underlined)

    @staticmethod
    def print_rgb(text: str, text_rgb: tuple[int, int, int], bg_rgb: tuple[int, int, int] | None = None,
                  bold: bool = False, underlined: bool = False) -> None:
        """Prints a message to the console."""
        formatted_text = text if Printer._formatted(text) else Printer.format_rgb(text, text_rgb, bg_rgb, bold,
                                                                                  underlined)
        print(formatted_text, end='\n' if Printer._newline else '', flush=True)

    @staticmethod
    def rgb_to_hex(rgb: tuple[int, int, int]) -> str:
        """Converts an RGB tuple to a hex string."""
        return f'#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}'

    @staticmethod
    def short_hex_to_long_hex(hex_color: str) -> str:
        """Converts a short hex color to a long hex color."""
        hex_without_hash = hex_color.replace('#', '') if hex_color.startswith('#') else hex_color
        return f'#{hex_without_hash[0]}{hex_without_hash[0]}{hex_without_hash[1]}{hex_without_hash[1]}' \
               f'{hex_without_hash[2]}{hex_without_hash[2]}'

    @staticmethod
    def success(text: str, bold: bool = False, underlined: bool = False) -> None:
        """Prints a success message to the console."""
        Printer.print_hex(text, TerminalColors.SUCCESS, bold=bold, underlined=underlined)

    @staticmethod
    def strip_ansi(text: str) -> str:
        """Strips ANSI escape sequences from a string."""
        return re.sub(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])', '', text)

    @staticmethod
    def strip_colors(text: str) -> str:
        stripped_text = re.sub(r'\x1b[\[\d;]+m', '', text)
        return Printer.strip_ansi(stripped_text)

    @staticmethod
    def warning(text: str, bold: bool = False, underlined: bool = False) -> None:
        """Prints a warning message to the console."""
        Printer.print_hex(text, TerminalColors.WARNING, bold=bold, underlined=underlined)

    @staticmethod
    def _formatted(text: str) -> bool:
        return str(text).endswith(TerminalColors.END_CHAR)



from .progress_renderer_args import LabelRendererArgs, PercentageRendererArgs, ProgressBarRendererArgs, \
    SpinnerRendererArgs
from .progress import Progress
from ..printer import Printer


class ColorfulProgress(Progress):
    def __init__(self, start: int = 0, end: int = 100, value: int = 0):
        super().__init__(start, end, value)
        self._setup_renderers()

    @staticmethod
    def get_basic_colored_text(text: str, percentage: float):
        if percentage < 15:
            return Printer.format_hex(text, '#FF0D0D')
        elif percentage < 30:
            return Printer.format_hex(text, '#FF4E11')
        elif percentage < 45:
            return Printer.format_hex(text, '#FF8E15')
        elif percentage < 60:
            return Printer.format_hex(text, '#FAB733')
        elif percentage < 75:
            return Printer.format_hex(text, '#ACB334')
        elif percentage < 90:
            return Printer.format_hex(text, '#69B34C')
        else:
            return Printer.format_hex(text, '#0EB33B')

    @staticmethod
    def _label_renderer(args: LabelRendererArgs):
        if args.percentage < 100:
            return Printer.format_hex(args.label, '#FFCC75')
        return Printer.format_hex(args.label, '#0EB33B')

    @staticmethod
    def _percentage_renderer(args: PercentageRendererArgs):
        return ColorfulProgress.get_basic_colored_text("{:.2f}%".format(args.percentage), args.percentage)

    @staticmethod
    def _progress_renderer(args: ProgressBarRendererArgs):
        return ColorfulProgress.get_basic_colored_text(args.text, args.percentage)

    @staticmethod
    def _spinner_renderer(args: SpinnerRendererArgs):
        return ColorfulProgress.get_basic_colored_text(args.spinner, args.percentage)

    def _setup_renderers(self):
        self.label_renderer = lambda args: self._label_renderer(args)  # LabelRendererArgs
        self.percentage_renderer = lambda args: self._percentage_renderer(args)  # PercentageRendererArgs
        self.progress_bar_renderer = lambda args: self._progress_renderer(args)  # ProgressBarRendererArgs
        self.spinner_renderer = lambda args: self._spinner_renderer(args)  # SpinnerRendererArgs
        self.info_text_renderer = lambda args: self._info_text_renderer(args.text)  # InfoTextRendererArgs
        self.info_separator_renderer = lambda args: self._info_separator_renderer()  # InfoSeparatorRendererArgs



import os


class FileUtils:
    @staticmethod
    def get_readable_file_size(size: int, suffix: str = "B"):
        for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
            if abs(size) < 1024.0:
                return "%3.2f%s%s%s" % (size, " ", unit, suffix)
            size /= 1024.0
        return "%.1f%s%s%s" % (size, " ", 'Y', suffix)

    @staticmethod
    def list_files(path, recursive=False, fullpath=True):
        if recursive:
            if not fullpath:
                filelist = FileUtils._list_files_recursively(path)
                return FileUtils._get_base_names_only(filelist)
            else:
                return FileUtils._list_files_recursively(path)
        else:
            if not fullpath:
                filelist = FileUtils._list_files_non_recursively(path)
                return FileUtils._get_base_names_only(filelist)
            else:
                return FileUtils._list_files_non_recursively(path)

    @staticmethod
    def list_folders(path, recursive=False, fullpath=True):
        if recursive:
            if not fullpath:
                folder_list = FileUtils._list_folders_recursively(path)
                return FileUtils._get_base_names_only(folder_list)
            else:
                return FileUtils._list_folders_recursively(path)
        else:
            if not fullpath:
                folder_list = FileUtils._list_folders_non_recursively(path)
                return FileUtils._get_base_names_only(folder_list)
            else:
                return FileUtils._list_folders_non_recursively(path)

    @staticmethod
    def _list_files_recursively(path):
        return [os.path.join(dp, f) for dp, dn, filenames in os.walk(path) for f in filenames]

    @staticmethod
    def _list_files_non_recursively(path):
        return [os.path.join(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]

    @staticmethod
    def _get_base_names_only(filelist: list):
        return [os.path.basename(f) for f in filelist]

    @staticmethod
    def _list_folders_recursively(path):
        return [dp for dp, dn, filenames in os.walk(path) if dp != path]

    @staticmethod
    def _list_folders_non_recursively(path):
        return [os.path.join(path, item) for item in os.listdir(path) if os.path.isdir(os.path.join(path, item))]
